#ifndef I2C_EXPORT_H
#define I2C_EXPORT_H

/****************************************************
 * $Id: i2c_export.h,v 1.1 2000/11/20 17:22:33 wdenk Exp $
 *
 * Copyright Motrola 1999
 *
 * $Log: i2c_export.h,v $
 * Revision 1.1  2000/11/20 17:22:33  wdenk
 *
 * * Added support for MBX860T (thanks to Rob Taylor)
 *
 * * Added support for Sandpoint8240 (thanks to Rob Taylor); this is
 *   Work In Progress (TM); current status: boots to command line input.
 *   EPIC code non-functional (interrupts disabled), No net, No IDE.
 *
 * * Add support for Status LED
 *
 * * Optionally panic() to reboot instead of hanging
 *
 * * Misc bug fixes
 *
 * * All frequencies in HZ now (internally)
 *
 * * Add support for BOOTP Domain Name Server Option
 *
 * Revision 1.1.1.1  2000/11/14 16:54:07  robt
 * no message
 *
 * Revision 1.1.1.1  2000/11/14 14:49:39  robt
 * no message
 *
 * Revision 1.3  1999/05/03 14:57:54  charliem
 * Decouple the I2C library from DINK application source by making the PRINT
 * function an optional parameter that is passed to I2C_Inititalize.  If the
 * function pointer is passed, it is used by the library; otherwise, no print
 * output is performed.
 *
 * Revision 1.2  1999/04/30 16:25:53  charliem
 * Change the DINK to I2C_do_transaction interface to allow the application
 * to specify the data address as a parameter to I2C_do_transaction, rather
 * than putting the data address into the first byte of the buffer.
 * There is a work-around in I2C_do_transaction in this version, see comments
 * in the i2c1.c code.
 *
 * Revision 1.1  1999/04/29 13:41:07  charliem
 * Capture first pass at modifications to the I2C interface.  This version is
 * correct, but there are some additional modifications to be implemented.
 * Nothing but I2C should be affected.
 *
 *
 ****************************************************/

/* These are the defined return values for the I2C_do_transaction function.
 * Any non-zero value indicates failure.  Failure modes can be added for 
 * more detailed error reporting.
 */
typedef enum _i2c_status
{
 I2C_SUCCESS     = 0,
 I2C_ERROR,
} I2C_Status;

/* These are the defined tasks for I2C_do_transaction.
 * Modes for SLAVE_RCV and SLAVE_XMIT will be added.
 */
typedef enum _i2c_transaction_mode
{
	I2C_MASTER_RCV =  0,
	I2C_MASTER_XMIT = 1,
} I2C_TRANSACTION_MODE;

typedef enum _i2c_interrupt_mode
{
	I2C_INT_DISABLE =  0,
	I2C_INT_ENABLE = 1,
} I2C_INTERRUPT_MODE;

typedef enum _i2c_stop
{
	I2C_NO_STOP =  0,
	I2C_STOP = 1,
} I2C_STOP_MODE;

typedef enum _i2c_restart
{
	I2C_NO_RESTART =  0,
	I2C_RESTART = 1,
} I2C_RESTART_MODE;

/******************** App. API ********************
 * The application API is for user level application 
 * to use the functionality provided by I2C driver.
 * This is a "generic" I2C interface, it should contain
 * nothing specific to the Kahlua implementation.
 * Only the generic functions are exported by the library.
 * 
 * Note: Its App.s responsibility to swap the data
 *       byte. In our API, we just transfer whatever
 *       we are given
 **************************************************/


/*  Initialize I2C unit with the following:
 *  driver's slave address 
 *  interrupt enabled
 *  optional pointer to application layer print function
 *
 *  These parameters may be added:
 *  desired clock rate
 *  digital filter frequency sampling rate
 *
 *  This function must be called before I2C unit can be used.
 */
extern I2C_Status I2C_Initialize( 
	unsigned char addr,            /* driver's I2C slave address */
	I2C_INTERRUPT_MODE en_int,     /* 1 - enable I2C interrupt 
	                                * 0 - disable I2C interrupt
	                                */
        int (*app_print_function)(char *,...)); /* pointer to optional "printf"
                                                 * provided by application
                                                 */

/* Perform the given I2C transaction, only MASTER_XMIT and MASTER_RCV
 * are implemented.  Both are only in polling mode.
 * 
 * en_int controls interrupt/polling mode
 * act is the type of transaction
 * addr is the I2C address of the slave device
 * len is the length of data to send or receive
 * buffer is the address of the data buffer
 * stop = I2C_NO_STOP, don't signal STOP at end of transaction
 *        I2C_STOP, signal STOP at end of transaction
 * retry is the timeout retry value, currently ignored
 * rsta = I2C_NO_RESTART, this is not continuation of existing transaction
 *        I2C_RESTART, this is a continuation of existing transaction
 */
extern I2C_Status I2C_do_transaction( I2C_INTERRUPT_MODE en_int,
                                      I2C_TRANSACTION_MODE act,
                                      unsigned char i2c_addr,
                                      unsigned char data_addr,
                                      int len,
                                      char *buffer,
                                      I2C_STOP_MODE stop,
                                      int retry,
                                      I2C_RESTART_MODE rsta);
#endif
